/*
 * Copyright (c) 2021 RidgeRun, LLC.
 * GPL-v2 License
 * Adapted by: Luis Leon <luis.leon@ridgerun.com>
 */

/**
 * SECTION:element-gstself
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch-1.0 videotestsrc ! "video/x-raw,width=640,height=480" ! \
 *                cudapassthru ! videoconvert ! ximagesink
 * ]|
 * Copy buffers in GstVideoFrame using OpenCV and transmit them
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstcudapassthru.h"

#include <EGL/egl.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <gst/gst.h>
#include <gst/video/gstvideofilter.h>
#include <gst/video/video.h>

#include <iostream>

#include "utils.h"

GST_DEBUG_CATEGORY_STATIC (gst_cuda_passthru_debug_category);

#define GST_CAT_DEFAULT gst_cuda_passthru_debug_category

struct _GstCudaPassThru
{
  GstVideoFilter base;

  CUcontext ccontext;
  cudaStream_t cstream;
  EGLDisplay egl_display;
};

/* prototypes */

static void gst_cuda_passthru_dispose (GObject * object);
static void gst_cuda_passthru_set_property (GObject * object, guint property_id,
    const GValue * value, GParamSpec * pspec);
static void gst_cuda_passthru_get_property (GObject * object, guint property_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_cuda_passthru_start (GstBaseTransform * trans);
static gboolean gst_cuda_passthru_stop (GstBaseTransform * trans);

static GstFlowReturn gst_cuda_passthru_transform_frame_ip (GstVideoFilter *
    filter, GstVideoFrame * frame);

enum
{
  PROP_0,
};

/* pad templates */

#define VIDEO_SRC_CAPS GST_VIDEO_CAPS_MAKE("BGR")

#define VIDEO_SINK_CAPS GST_VIDEO_CAPS_MAKE("BGR")

/* class initialization */

G_DEFINE_TYPE_WITH_CODE (GstCudaPassThru, gst_cuda_passthru,
    GST_TYPE_VIDEO_FILTER,
    GST_DEBUG_CATEGORY_INIT (gst_cuda_passthru_debug_category, "cudapassthru",
        0, "debug category for cudapassthru element"));

static void
gst_cuda_passthru_class_init (GstCudaPassThruClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstBaseTransformClass *base_transform_class =
      GST_BASE_TRANSFORM_CLASS (klass);
  GstVideoFilterClass *video_filter_class = GST_VIDEO_FILTER_CLASS (klass);

  /* Setting up pads and setting metadata should be moved to
     base_class_init if you intend to subclass this class. */
  gst_element_class_add_pad_template (GST_ELEMENT_CLASS (klass),
      gst_pad_template_new ("src", GST_PAD_SRC, GST_PAD_ALWAYS,
          gst_caps_from_string (VIDEO_SRC_CAPS)));
  gst_element_class_add_pad_template (GST_ELEMENT_CLASS (klass),
      gst_pad_template_new ("sink", GST_PAD_SINK, GST_PAD_ALWAYS,
          gst_caps_from_string (VIDEO_SINK_CAPS)));

  gobject_class->set_property = gst_cuda_passthru_set_property;
  gobject_class->get_property = gst_cuda_passthru_get_property;

  gst_element_class_set_static_metadata (GST_ELEMENT_CLASS (klass),
      "Cuda Pass-Thru element", "Generic",
      "Cuda Pass-Thru element over a video stream", "<luis.leon@ridgerun.com>");

  /* Properties setup */

  base_transform_class->start = GST_DEBUG_FUNCPTR (gst_cuda_passthru_start);
  base_transform_class->stop = GST_DEBUG_FUNCPTR (gst_cuda_passthru_stop);
  gobject_class->dispose = GST_DEBUG_FUNCPTR (gst_cuda_passthru_dispose);
  video_filter_class->transform_frame_ip =
      GST_DEBUG_FUNCPTR (gst_cuda_passthru_transform_frame_ip);
}

static void
gst_cuda_passthru_init (GstCudaPassThru * self)
{
  CUdevice device;

  /* Initialise EGL-related things */
  self->egl_display = eglGetDisplay (EGL_DEFAULT_DISPLAY);
  if (self->egl_display == EGL_NO_DISPLAY) {
    GST_ERROR_OBJECT (self, "Could not get EGL display (try 'unset DISPLAY')");
    goto done;
  }

  if (!eglInitialize (self->egl_display, NULL, NULL)) {
    GST_ERROR_OBJECT (self,
        "Could not initialize EGL display (try 'unset DISPLAY')");
    goto done;
  }

  /* Initialise CUDA-related things */
  CU_JUMP_IF_FAIL (done, cuInit (0));
  CU_JUMP_IF_FAIL (done, cuDeviceGet (&device, 0));
  CU_JUMP_IF_FAIL (done, cuCtxCreate (&self->ccontext, 0, device));

  CUDA_JUMP_IF_FAIL (done, cudaFree (0));
  CUDA_JUMP_IF_FAIL (done, cudaStreamCreateWithFlags (&self->cstream,
          cudaStreamNonBlocking));

  /* Other stuff */

  /* Synchronise */
  CUDA_JUMP_IF_FAIL (done, cudaStreamSynchronize (self->cstream));
  CU_JUMP_IF_FAIL (done, cuCtxSynchronize ());

done:
  GST_DEBUG_OBJECT (self, "Init...");
}

void
gst_cuda_passthru_set_property (GObject * object, guint property_id,
    const GValue * value, GParamSpec * pspec)
{
  GstCudaPassThru *self = GST_CUDA_PASSTHRU (object);

  GST_DEBUG_OBJECT (self, "set_property");

  GST_OBJECT_LOCK (self);
  switch (property_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
  }
  GST_OBJECT_UNLOCK (self);
}

void
gst_cuda_passthru_get_property (GObject * object, guint property_id,
    GValue * value, GParamSpec * pspec)
{
  GstCudaPassThru *self = GST_CUDA_PASSTHRU (object);

  GST_DEBUG_OBJECT (self, "get_property");

  GST_OBJECT_LOCK (self);
  switch (property_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
  }
  GST_OBJECT_UNLOCK (self);
}

static gboolean
gst_cuda_passthru_start (GstBaseTransform * trans)
{
  GstCudaPassThru *self = GST_CUDA_PASSTHRU (trans);

  GST_INFO_OBJECT (self, "Starting");

  return TRUE;
}

static gboolean
gst_cuda_passthru_stop (GstBaseTransform * trans)
{
  GstCudaPassThru *self = GST_CUDA_PASSTHRU (trans);

  GST_INFO_OBJECT (self, "Stopping");

  return TRUE;
}

static GstFlowReturn
gst_cuda_passthru_transform_frame_ip (GstVideoFilter * filter,
    GstVideoFrame * frame)
{
  GstCudaPassThru *self = GST_CUDA_PASSTHRU (filter);

  GST_INFO_OBJECT (self, "Transforming");
  return GST_FLOW_OK;
}

static void
gst_cuda_passthru_dispose (GObject * object)
{
  GstCudaPassThru *self = GST_CUDA_PASSTHRU (object);

  // CUDA Cleanup
  if (self->cstream) {
    (void) cudaStreamDestroy (self->cstream);
    self->cstream = NULL;
  }

  if (self->ccontext) {
    (void) cuCtxDestroy (self->ccontext);
    self->ccontext = NULL;
  }

  if (self->egl_display != EGL_NO_DISPLAY) {
    (void) eglTerminate (self->egl_display);
  }

  G_OBJECT_CLASS (gst_cuda_passthru_parent_class)->dispose (object);
}
