/*
 * Copyright (c) 2021 RidgeRun, LLC.
 * GPL-v2 License
 * Adapted by: Luis Leon <luis.leon@ridgerun.com>
 */

/**
 * SECTION:element-gstself
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch-1.0 videotestsrc ! "video/x-raw,width=640,height=480" ! \
 *                torchpassthru ! videoconvert ! ximagesink
 * ]|
 * Copy buffers in GstVideoFrame using OpenCV and transmit them
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gst/gst.h>
#include <gst/video/gstvideofilter.h>
#include <gst/video/video.h>
#include <torch/torch.h>
#include <c10/cuda/CUDAStream.h>
#include <iostream>

#include "gsttorchpassthru.h"


GST_DEBUG_CATEGORY_STATIC (gst_torch_passthru_debug_category);

#define GST_CAT_DEFAULT gst_torch_passthru_debug_category


struct _GstTorchPassThru
{
  GstVideoFilter base;
};

/* prototypes */


static void gst_torch_passthru_set_property (GObject * object,
    guint property_id, const GValue * value, GParamSpec * pspec);
static void gst_torch_passthru_get_property (GObject * object,
    guint property_id, GValue * value, GParamSpec * pspec);

static gboolean gst_torch_passthru_start (GstBaseTransform * trans);
static gboolean gst_torch_passthru_stop (GstBaseTransform * trans);

static GstFlowReturn gst_torch_passthru_transform_frame_ip (GstVideoFilter * filter,
    GstVideoFrame * frame);

enum
{
  PROP_0,
};


/* pad templates */

#define VIDEO_SRC_CAPS \
    GST_VIDEO_CAPS_MAKE("BGR")

#define VIDEO_SINK_CAPS \
   GST_VIDEO_CAPS_MAKE("BGR")


/* class initialization */

G_DEFINE_TYPE_WITH_CODE (GstTorchPassThru, gst_torch_passthru, GST_TYPE_VIDEO_FILTER,
    GST_DEBUG_CATEGORY_INIT (gst_torch_passthru_debug_category, "torchpassthru", 0,
        "debug category for torchpassthru element"));

static void
gst_torch_passthru_class_init (GstTorchPassThruClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstBaseTransformClass *base_transform_class =
      GST_BASE_TRANSFORM_CLASS (klass);
  GstVideoFilterClass *video_filter_class = GST_VIDEO_FILTER_CLASS (klass);

  /* Setting up pads and setting metadata should be moved to
     base_class_init if you intend to subclass this class. */
  gst_element_class_add_pad_template (GST_ELEMENT_CLASS (klass),
      gst_pad_template_new ("src", GST_PAD_SRC, GST_PAD_ALWAYS,
          gst_caps_from_string (VIDEO_SRC_CAPS)));
  gst_element_class_add_pad_template (GST_ELEMENT_CLASS (klass),
      gst_pad_template_new ("sink", GST_PAD_SINK, GST_PAD_ALWAYS,
          gst_caps_from_string (VIDEO_SINK_CAPS)));

  gobject_class->set_property = gst_torch_passthru_set_property;
  gobject_class->get_property = gst_torch_passthru_get_property;


  gst_element_class_set_static_metadata (GST_ELEMENT_CLASS (klass),
      "Torch Pass-Thru element", "Generic",
      "Torch Pass-Thru element over a video stream",
      "<luis.leon@ridgerun.com>");


  /* Properties setup */

  base_transform_class->start = GST_DEBUG_FUNCPTR (gst_torch_passthru_start);
  base_transform_class->stop = GST_DEBUG_FUNCPTR (gst_torch_passthru_stop);
  video_filter_class->transform_frame_ip =
      GST_DEBUG_FUNCPTR (gst_torch_passthru_transform_frame_ip);
}

static void
gst_torch_passthru_init (GstTorchPassThru * self)
{
}

void
gst_torch_passthru_set_property (GObject * object, guint property_id,
    const GValue * value, GParamSpec * pspec)
{
  GstTorchPassThru *self = GST_TORCH_PASSTHRU (object);

  GST_DEBUG_OBJECT (self, "set_property");

  GST_OBJECT_LOCK(self);
  switch (property_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
  }
  GST_OBJECT_UNLOCK(self);

}

void
gst_torch_passthru_get_property (GObject * object, guint property_id,
    GValue * value, GParamSpec * pspec)
{
  GstTorchPassThru *self = GST_TORCH_PASSTHRU (object);

  GST_DEBUG_OBJECT (self, "get_property");

  GST_OBJECT_LOCK(self);
  switch (property_id) {
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
  }
  GST_OBJECT_UNLOCK(self);
}


static gboolean
gst_torch_passthru_start (GstBaseTransform * trans)
{
  GstTorchPassThru *self = GST_TORCH_PASSTHRU (trans);

  GST_INFO_OBJECT (self, "Starting");

  return TRUE;
}

static gboolean
gst_torch_passthru_stop (GstBaseTransform * trans)
{
  GstTorchPassThru *self = GST_TORCH_PASSTHRU (trans);

  GST_INFO_OBJECT (self, "Stopping");

  return TRUE;
}


//#define USE_CUDA
//#define USE_TENSOR_OP
//#define USE_STREAMS

static GstFlowReturn
gst_torch_passthru_transform_frame_ip (GstVideoFilter * filter,
    GstVideoFrame * frame)
{
  GstTorchPassThru *self = GST_TORCH_PASSTHRU (filter);
#ifdef USE_TENSOR_OP
#ifdef USE_CUDA
  torch::Tensor tensor1 = torch::eye(3, torch::device(torch::kCUDA));
  torch::Tensor tensor2 = torch::eye(3, torch::device(torch::kCUDA));
#else
  torch::Tensor tensor1 = torch::eye(3);
  torch::Tensor tensor2 = torch::eye(3);
#endif

#ifdef USE_STREAMS
  at::cuda::CUDAStream myStream = at::cuda::getStreamFromPool();
  at::cuda::setCurrentCUDAStream(myStream);
#endif
  torch::Tensor tensor3 = tensor1 + tensor2;

  std::cout << tensor1 << "+" << tensor2 << "=" << tensor3 << std::endl;

#ifdef USE_STREAMS
  at::cuda::CUDAStream defaultStream = at::cuda::getDefaultCUDAStream();
  at::cuda::setCurrentCUDAStream(defaultStream);
#endif
#endif

  GST_INFO_OBJECT (self, "Transforming");
  return GST_FLOW_OK;
}
