/*
 * Copyright (c) 2021 RidgeRun, LLC.
 * GPL-v2 License
 * Adapted by: Luis Leon <luis.leon@ridgerun.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>

#include <glib.h>
#include <gst/gst.h>

#define NUM_PROCS 2
#define NUM_ENDS 2
#define MSGSIZE 20

static gint meshinator(gint pipes[NUM_ENDS], gint pid);
static gint detection(gint pipes[NUM_ENDS], gint pid);

int main(int argc, char **argv) {
  gint pid = 0;
  gint ret = 0;
  gint pipes[NUM_ENDS];

  if (pipe(pipes) < 0) exit(1);

  /* Initalisation before the fork. Maybe it should be after... */
  gst_init(&argc, &argv);

  pid = fork();
  if (0 == pid) {
    /* Master */
    ret = detection(pipes, pid);
  } else {
    /* Child */
    ret = meshinator(pipes, pid);
    /* Close */
    close(pipes[1]);
  }

  wait(NULL);
  return ret;
}

static gboolean bus_call(GstBus *bus, GstMessage *msg, gpointer data) {
  GMainLoop *loop = (GMainLoop *)data;

  switch (GST_MESSAGE_TYPE(msg)) {
    case GST_MESSAGE_EOS:
      g_print("End of stream\n");
      g_main_loop_quit(loop);
      break;

    case GST_MESSAGE_ERROR: {
      gchar *debug;
      GError *error;

      gst_message_parse_error(msg, &error, &debug);
      g_free(debug);

      g_printerr("Error: %s\n", error->message);
      g_error_free(error);

      g_main_loop_quit(loop);
      break;
    }
    default:
      break;
  }

  return TRUE;
}

void handoff_signal(GstElement *object, GstBuffer *arg0, GstPad *arg1,
                    gpointer user_data) {
  g_return_if_fail(user_data);
  g_print("Handoff signal received. PID: %i\n", *((gint *)(user_data)));
}

#define MESHINATOR_PIPE                                                     \
  "videotestsrc pattern=black is-live=1 num-buffers=1000 ! nvvidconv ! "    \
  "d3meshinator "                                                           \
  "filename=/home/nvidia/D3/001-camerasrc/d3meshinator/eg/mod1_Cal.json ! " \
  "fakesink "                                                               \
  "name=mysink signal-handoffs=true sync=false"

static gint meshinator(gint pipes[NUM_ENDS], gint pid) {
  GMainLoop *loop = NULL;
  GstElement *pipeline = NULL, *sink = NULL;
  GstBus *bus = NULL;
  guint bus_watch_id = 0;
  GError *error = NULL;
  char msg1[MSGSIZE] = "Start Detector";
  char inbuf[MSGSIZE];

  loop = g_main_loop_new(NULL, FALSE);
  pipeline = gst_parse_launch(MESHINATOR_PIPE, &error);

  if (error) {
    g_printerr("Error parsing pipeline %s: %s", MESHINATOR_PIPE,
               error->message);
    g_error_free(error);
    return -1;
  }

  bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
  bus_watch_id = gst_bus_add_watch(bus, bus_call, loop);
  gst_object_unref(bus);

  /* Connect signal */
  sink = gst_bin_get_by_name(GST_BIN(pipeline), "mysink");
  g_signal_connect(sink, "handoff", G_CALLBACK(handoff_signal), &pid);

  gst_element_set_state(pipeline, GST_STATE_PLAYING);

  /* Receive signal from the detector */
  g_print("Sending start signal...\n");
  read(pipes[0], inbuf, MSGSIZE);
  g_print("Starting meshinator...\n");

  /* Run the loop */
  g_main_loop_run(loop);
  gst_element_set_state(pipeline, GST_STATE_NULL);

  g_print("Deleting pipeline...\n");

  gst_object_unref(GST_OBJECT(pipeline));
  g_source_remove(bus_watch_id);
  g_main_loop_unref(loop);

  return 0;
}

#define DETECTION_PIPE                                                 \
  "videotestsrc pattern=black is-live=1 num-buffers=1000 ! "           \
  "video/x-raw,width=640,height=480 ! torchpassthru ! videoconvert ! " \
  "fakesink name=mysink "                                              \
  "signal-handoffs=true sync=false"

static gint detection(gint pipes[NUM_ENDS], gint pid) {
  GMainLoop *loop = NULL;
  GstElement *pipeline = NULL, *sink = NULL;
  GstBus *bus = NULL;
  guint bus_watch_id = 0;
  char msg1[MSGSIZE] = "Start Meshinator";
  char inbuf[MSGSIZE];
  GError *error = NULL;

  loop = g_main_loop_new(NULL, FALSE);
  pipeline = gst_parse_launch(DETECTION_PIPE, &error);

  if (error) {
    g_printerr("Error parsing pipeline %s: %s", DETECTION_PIPE, error->message);
    g_error_free(error);
    return -1;
  }

  bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
  bus_watch_id = gst_bus_add_watch(bus, bus_call, loop);
  gst_object_unref(bus);

  /* Connect signal */
  sink = gst_bin_get_by_name(GST_BIN(pipeline), "mysink");
  g_signal_connect(sink, "handoff", G_CALLBACK(handoff_signal), &pid);

  gst_element_set_state(pipeline, GST_STATE_PLAYING);

  /* Send signal to the meshinators to start*/
  g_print("Sending mesh start signal...\n");
  write(pipes[1], msg1, MSGSIZE);
  g_print("Starting detector...\n");

  /* Run the loop */
  g_main_loop_run(loop);
  gst_element_set_state(pipeline, GST_STATE_NULL);

  g_print("Deleting pipeline...\n");

  gst_object_unref(GST_OBJECT(pipeline));
  g_source_remove(bus_watch_id);
  g_main_loop_unref(loop);

  return 0;
}
